import abc
import argparse
import importlib
import logging
import os
import random
from random import sample
import re
from subprocess import Popen
import shlex
import sys
import time
from threading import Thread, Lock
import traceback
import functools
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.animation as animation
import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
from torch.autograd import Variable
from py4j.java_gateway import JavaGateway
from py4j.java_gateway import GatewayParameters
from py4j.java_gateway import CallbackServerParameters
from py4j.java_gateway import get_field
from py4j.java_gateway import set_field
try:
    from tqdm import trange
except ImportError as exc:
    def _trange(n, *args, **kwargs):
        logger.info('Install "tqdm" to see progress bar. '
                    'Type "pip install tqdm" in terminal')
        return range(n)
    trange = _trange


logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
logger.addHandler(logging.StreamHandler())
# logger can printout message from BasicBot(AIInterface)
# Because, Fighting Game Platform block print's functionality,
# We can't see error message, also can't see print message for debugging
# logging module is good solution for this case


FTG_PATH = '../FightingICEver.3.10'
bot_name = __file__[:-3]  # BasicBot
log_file = '{}.csv'.format(bot_name)
model_file = '{}.pt'.format(bot_name)
root_path = '.'

resolution = (4, 38, 57)
# (channel, height, width),
# channel: frames[t-3, t-2, t-1, t-0]
memory_size = 500000
batch_size = 32
learning_rate = 0.0001
gamma = 0.99  # discount factor
epsilon = 1.0  # initial epsilon (exploration rate) for epsilon greedy algorithm
reward_scale = 30   # scaled_reward = reward / reward_scale
energy_scale = 300  # scaled_energy = energy / energy_scale
frames = 4


# actions that BasicBot can perform
actions = [
    'LLLL',
    'RRRR',
    'B B ',
]


class ReplayMemory:
    def __init__(self, shape, capacity):
        self.lock = Lock()  # for multiprocessing, threading
        state_shape = (capacity, shape[0], shape[1], shape[2])
        self.s1 = np.zeros(state_shape, dtype=np.float32)
        self.s2 = np.zeros(state_shape, dtype=np.float32)
        self.a = np.zeros(capacity, dtype=np.int32)
        self.r = np.zeros(capacity, dtype=np.float32)
        self.done = np.zeros(capacity, dtype=np.float32)  # do not use this platform
        self.energy = np.zeros(capacity, dtype=np.float32)

        self.capacity = capacity
        self.size = 0
        self.pos = 0

    def add(self, s1, action, s2, done, reward, energy):
        with self.lock:
            self.s1[self.pos, :, :, :] = s1
            self.a[self.pos] = action
            if not done:
                self.s2[self.pos, :, :, :] = s2
            self.done[self.pos] = done
            self.r[self.pos] = reward
            self.energy[self.pos] = energy

            self.pos = (self.pos + 1) % self.capacity
            self.size = min(self.size + 1, self.capacity)

    def sample(self, sample_size):
        i = sample(range(0, self.size), sample_size)
        with self.lock:
            return self.s1[i], self.a[i], self.s2[i], self.done[i], self.r[i], self.energy[i]


class Net(nn.Module):
    def __init__(self, input_shape, n_outputs):
        super().__init__()
        channel, height, width = input_shape
        # [-1, 4, 38, 57]
        self.conv1 = nn.Conv2d(channel, 8, 6, stride=3)
        # [-1, 8, 11, 18]
        self.conv2 = nn.Conv2d(8, 8, 3, stride=2)
        # [-1, 8, 5, 8]
        self.conv2_drop = nn.Dropout2d()
        self.fc1 = nn.Linear(8 * 5 * 8 + 1, 128)
        self.fc2 = nn.Linear(128, n_outputs)

    def forward(self, s1, eng):
        try:
            s1_org = s1
            s1 = F.relu(self.conv1(s1_org))
            s1 = F.relu(self.conv2_drop(self.conv2(s1)))
            s1 = s1.view(-1, 8 * 5 * 8)
            x = torch.cat((s1, eng), 1)
            x = F.relu(self.fc1(x))
            x = F.dropout(x, training=self.training)
            x = self.fc2(x)
            return x
        except Exception as exc:
            logger.error(exc)
            exit(-1)


class Singleton(abc.ABCMeta):
    __instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls.__instances:
            cls.__instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls.__instances[cls]


LearnerLock = Lock()
net = Net(resolution, len(actions)).cuda()
optimizer = optim.RMSprop(net.parameters(), lr=learning_rate)
# optimizer = optim.SGD(net.parameters(), lr=learning_rate, momentum=0.05)
loss_func = torch.nn.MSELoss()
memory = ReplayMemory(resolution, memory_size)


class Coach(metaclass=Singleton):
    # - Do all jobs using neural network (learning, decision making, save and load parameters)
    # - Singleton: every objects can access this object, and others through this object (act, memorize)
    # - change training, testing mode
    def __init__(self):
        with LearnerLock:
            self.lock = LearnerLock
            # every BasicBot objects
            self.trainees = list()

            self.reward_log = [0]

    def save(self, model_file='BasicBot.pt', **kwargs):
        # save model's parameters
        # there are two files to save
        # - model.pt: saved in current working directory or root_path, current highest performance,
        #   save method call when new high performance archived
        # - model.pt-{id}: to preserve past best parameters, if pass over mean score (kwargs[hint]) from caller
        #   {id} == current score
        with self.lock:
            # save "./model.pt"
            model_path = os.path.join(root_path, model_file)
            torch.save(net.state_dict(), model_path)
            logger.info('Model file({}) saved'.format(model_path))

            # save backup "./models/model.pt-{id}"
            os.makedirs(os.path.join(root_path, 'models'), exist_ok=True)
            # {id}: current score (the highest score or current time
            model_path_aux = os.path.join(
                root_path, 'models', '{}-{}'.format(
                    model_file, kwargs.get('hint', time.ctime())))
            torch.save(net.state_dict(), model_path_aux)
            logger.info('Model file aux({}) saved'.format(model_path_aux))

    def load(self, model_file=None):
        if model_file is None:
            model_file = '{}.pt'.format(bot_name)

        paths = [os.path.join(FTG_PATH, 'data/aiData', bot_name, model_file),
                 os.path.join(root_path, model_file)]

        model_path = None
        for path in paths:
            if os.path.exists(path):
                model_path = path
                break

        with self.lock:
            if os.path.exists(model_path):
                net.load_state_dict(torch.load(model_path))
                logger.info('Model file({}) loaded'.format(model_path))
            else:
                logger.warning('Model file dose not exixts, '
                               'please place model file (BasicBot.pt) in current working directory '
                               'or {}/data/aiData/BasicBot'.format(FTG_PATH))

    def act(self, state, eps):
        # this method override all BasicBot objects' act
        with self.lock:  # every BasicBot objects can call this
            screens, energy = state

            if random.random() < eps:
                # exploration
                logger.debug('select random action')
                return random.randint(0, len(actions)-1)
            else:
                # exploitation
                logger.debug('select best action')
                screens = screens.reshape([1, resolution[0], resolution[1], resolution[2]])
                screens = Variable(torch.from_numpy(screens)).cuda()
                energy = Variable(torch.FloatTensor([[energy]])).cuda()
                q_values = net(screens, energy)
                return q_values.data.max(1)[1][0][0]

    def memorize(self, s1, a, s2, done, r, energy):
        # this method override all BasicBot objects' memorize
        memory.add(s1, a, s2, done, r, energy)

    def learn(self):
        s1, a, s2, _, r, energy = memory.sample(batch_size)
        s1 = Variable(torch.from_numpy(s1)).cuda()
        s2 = Variable(torch.from_numpy(s2)).cuda()
        r /= reward_scale
        energy = energy.reshape((batch_size, 1))
        energy = Variable(torch.from_numpy(energy)).cuda()

        with self.lock:  # get lock when only change parameters
            q1 = net(s1, energy).data.cpu().numpy()
            max_q = np.max(q1, axis=1)
            max_q2 = net(s2, energy).data.max(1)[0].cpu().numpy().reshape(batch_size)

            q1[np.arange(batch_size), a] = r + gamma * max_q2

            optimizer.zero_grad()
            target = Variable(torch.from_numpy(q1).cuda())
            output = net(s1, energy)
            loss = loss_func(output, target)
            loss.backward()
            optimizer.step()
            logger.debug('learn one batch')
            return max_q, loss.data[0]

    def training(self, eps):
        # set training mode and broadcast to all BasicBot objects
        logger.debug('Set training mode')
        for trainee in self.trainees:
            trainee.act = functools.partial(Coach().act, eps=eps)
            trainee.memorize = Coach().memorize

    def testing(self):
        # set testing mode and broadcast to all BasicBot objects
        logger.debug('Set testing mode')
        for trainee in self.trainees:
            trainee.act = functools.partial(Coach().act, eps=-1)
            trainee.memorize = Coach().memorize

    def add_reward(self, reward):
        with self.lock:
            self.reward_log.append(reward)  # save rewards for calculate score for certain period

    def get_rewards_stat(self):
        # this method called when training or testing end
        # to calculate training score or testing score
        with self.lock:
            rewards = np.array(self.reward_log)
            self.reward_log.clear()
            self.reward_log.append(0)
            return rewards


class BasicBot(object):
    def __init__(self, gateway):
        self.gateway = gateway

        self.frames = frames
        self.frame_count = 0

        # save short period of data to make 4 channel data
        self.capacity = resolution[0] * 5
        self.screens = list()
        self.actions = list()
        self.hps = list()
        self.energy = list()
        self.controllable = list()

        # FTGEnv object set this
        # use this to avoid change API of this bot
        on_train = getattr(self, '_on_train', False)

        # when start game, subscribe self to Coach object
        # and override act, and memorise
        if on_train:
            Coach().trainees.append(self)
            self.act = functools.partial(Coach().act, eps=1.0)
            self.memorize = Coach().memorize
        else:
            Coach().load(model_file)
            self.act = functools.partial(Coach().act, eps=-1)  # always do best action, epsilon == -1
            self.memorize = lambda s1, a, s2, done, r, energy: None  # do nothing

    def getCharacter(self):
        # return "LUD"
        return "ZEN"

    def close(self):
        pass

    def initialize(self, gameData, player):
        self.inputKey = self.gateway.jvm.structs.Key()
        self.frameData = self.gateway.jvm.structs.FrameData()
        self.cc = self.gateway.jvm.commandcenter.CommandCenter()

        self.player = player
        self.gameData = gameData
        return 0

    def getInformation(self, frameData):
        self.frameData = frameData
        self.cc.setFrameData(self.frameData, self.player)

    def input(self):
        return self.inputKey

    def processing(self):
        try:
            # initialize when game start (first processing call?)
            if self.frameData.getEmptyFlag() or self.frameData.getRemainingTime() <= 0:
                self.isGameJustStarted = True
                self.screens.clear()
                self.actions.clear()
                self.hps.clear()
                self.energy.clear()
                self.controllable.clear()
                self.frame_count = self.frames
                return

            # if self.cc.getSkillFlag():
            #     self.inputKey = self.cc.getSkillKey()
            #     return

            logger.debug('{} {} {}'.format(get_field(self.inputKey, 'L'),
                                           get_field(self.inputKey, 'R'),
                                           get_field(self.inputKey, 'B')))

            # update state
            displayBuffer = self.frameData.getDisplayByteBufferAsBytes(resolution[2], resolution[1], True)
            # rescaled to [-1, 1] and change data type np.float32
            screen = np.frombuffer(displayBuffer, dtype=np.int8).reshape(resolution[1:]) / 127
            screen = screen.astype(np.float32)
            self.screens.append(screen)

            my_char = self.frameData.getMyCharacter(self.player)
            opp_char = self.frameData.getOpponentCharacter(self.player)
            self.hps.append((my_char.getHp(), opp_char.getHp()))
            self.energy.append(my_char.getEnergy() / energy_scale)  # energy scaled

            # set action
            action_continue = self.actions and self.actions[-1] is not None \
                              and self.frame_count < self.frames
            self.controllable.append(get_field(my_char, 'control'))
            controllable = self.controllable[-1]

            # if last action is set and running
            if action_continue or not controllable:
                # keep current action
                self.actions.append(self.actions[-1])
            else:
                # take new action
                action_idx = self.act(self._get_recent_state())
                self.actions.append(action_idx)
                self.frames = len(actions[action_idx])
                logger.debug('action_idx: {}'.format(action_idx))
                self.frame_count = 0

            # set key
            if self.actions and self.actions[-1] is not None:
                current_action = actions[self.actions[-1]]
                for action in actions:
                    if self.frame_count < len(current_action):
                        key = action[0]
                        set_field(self.inputKey, key, key == current_action[self.frame_count])
            else:
                self.inputKey.empty()

            self.save_transaction()  # store new state action tuple and stream to monitor
            self.frame_count += 1

        except Exception as exc:
            logger.error(traceback.format_exc())

    def save_transaction(self):
        # save state and action tuple using Coach's memorize
        try:
            if len(self.screens) > self.capacity:
                del self.screens[0]
                del self.actions[0]
                del self.hps[0]
                del self.energy[0]

            if len(self.screens) == self.capacity:
                assert(len(self.screens) == len(self.actions)
                       == len(self.hps) == len(self.energy))
                s1 = np.stack(
                    [self.screens[0],  # 1st frame
                     self.screens[4],  # 2nd frame
                     self.screens[8],  # 3rd frame
                     self.screens[12]])  # 4th frame
                s2 = np.stack(
                    [self.screens[4],  # 2nd frame
                     self.screens[8],  # 3rd frame
                     self.screens[12],  # 4th frame
                     self.screens[16]])  # 5th frame
                a = self.actions[0]
                # reward calculation
                my_hp_1, opp_hp_1 = self.hps[0]
                my_hp_2, opp_hp_2 = self.hps[4]
                r = (opp_hp_1 - opp_hp_2) - (my_hp_1 - my_hp_2)
                energy = self.energy[0]

                if self.controllable[0]:
                    # memorize sample when character is controllable
                    # do not use game end in this platform
                    # done is always false
                    self.memorize(s1, a, s2, False, r, energy)

                # for calculate score
                Coach().add_reward(r)

                # if there are debug port (process queue for monitoring code)
                # send current state and action  through this
                if hasattr(self, 'debug_port'):
                    cnt = getattr(self, '_debug_port_cnt', 0)
                    if cnt == 0:
                        screens, energy = self._get_recent_state()
                        info = dict(action=actions[self.actions[-1]],
                                    energy=int(energy * energy_scale),
                                    reward=r)
                        self.debug_port.put((screens, info))
                    setattr(self, '_debug_port_cnt', (cnt + 1) % 4)
        except Exception as exc:
            logger.error(traceback.format_exc())

    def _get_recent_state(self):
        if len(self.screens) == self.capacity:
            screen = np.stack([self.screens[-13],  # 1st frame
                               self.screens[-9],  # 2nd frame
                               self.screens[-5],  # 3rd frame
                               self.screens[-1]])  # 4st frame
        else:
            screen = np.stack([self.screens[-1],
                               self.screens[-1],
                               self.screens[-1],
                               self.screens[-1]])
        energy = self.energy[-1]
        return screen, energy

    # This part is mandatory
    class Java:
        implements = ["gameInterface.AIInterface"]


class FTGEnv(object):
    # launching Fighting Game AI platform as a separated process and manage it
    def __init__(self, _id, player_1, player_2, *args, **kwargs):
        self.id = _id
        self.train = kwargs.get('train', True)
        self.port = kwargs.get('port', 4242)
        self.black_bg = kwargs.get('black_bg', True)
        self.disable_window = kwargs.get('disable_window', True)
        self.verbose = kwargs.get('verbose', False)
        self.player_1 = player_1  # player 1 or 2 class
        self.player_2 = player_2  # not instance
        self.inverted_player_number = kwargs.get('inverted_player', 2)
        self.starts_with_energy = kwargs.get('starts_with_energy', False)

        self.log_off = kwargs.get('log_off', True)

        self.game_proc = None  # FTG game server process
        self.game_thread = None  # game client thread (belong to this process)
        self.manager = None
        self.gate_way = None

    def __del__(self):
        self.close()

    def close(self):
        if self.game_proc:
            logger.info('Terminate FTG {}'.format(self.id))
            self.game_proc.terminate()

    def _start_server(self, **kwargs):
        root_path = kwargs.get('root_path', '')
        cps = ['FightingICE.jar', 'lib/gameLib.jar', 'lib/lwjgl.jar',
               'lib/lwjgl_util.jar', 'lib/javatuples-1.2.jar', 'lib/commons_csv.jar',
               'lib/jinput.jar', 'lib/fileLib.jar', 'lib/py4j0.10.4.jar']
        cps += [os.path.join('data/ai', p) for p in os.listdir('data/ai') if p.endswith('.jar')]
        cps = [os.path.join(root_path, cp) for cp in cps]

        cmd = [
            'java',
            '-cp', ':'.join(cps),
            '-Djava.library.path="{}"'.format(os.path.join(root_path, 'lib/native/linux')),
            'Main',
            '--py4j',
        ]

        if self.black_bg:
            cmd.append('--black-bg')

        cmd.append('--inverted-player {}'.format(self.inverted_player_number))
        # cmd.append('--inverted-player 2')
        cmd.append('--port %d' % self.port)

        if self.disable_window:
            cmd.append("--disable-window")

        if self.log_off:
            cmd.append('--off')
            cmd.append('--del')

        if self.starts_with_energy:
            cmd.append('-t')

        if not self.verbose:
            server_log = open(os.devnull, 'w')
        else:
            import sys
            server_log = sys.stdout

        # FTG process start
        self.game_proc = Popen(shlex.split(' '.join(cmd)), stdout=server_log, stderr=server_log)
        time.sleep(3)

    def _create_player(self, gateway, manager, player):
        if isinstance(player, str):
            # java bot: string input
            player_name, player_character = player.split(':')
            player_obj = None
        else:
            # python bot: python class input
            player._on_train = self.train
            player_obj = player(gateway)
            player_name = player_obj.__class__.__name__
            manager.registerAI(player_name, player_obj)
            player_character = player_obj.getCharacter()
        return player_character, player_name, player_obj

    def _make_a_game(self, p1, p2, n_game, **kwargs):
        self.gateway = JavaGateway(gateway_parameters=GatewayParameters(port=self.port),
                                   callback_server_parameters=CallbackServerParameters(port=0), )
        python_port = self.gateway.get_callback_server().get_listening_port()
        self.gateway.java_gateway_server.resetCallbackClient(
            self.gateway.java_gateway_server.getCallbackClient().getAddress(), python_port)
        self.manager = self.gateway.entry_point

        connect_monitor = kwargs.get('ai_monitor', lambda a, b, c: None)

        def run_a_game():
            n = 0
            while n_game is None or n < n_game:
                p1_character, p1_name, p1_obj = self._create_player(self.gateway, self.manager, p1)
                connect_monitor(self.id, 1, p1_obj)
                p2_character, p2_name, p2_obj = self._create_player(self.gateway, self.manager, p2)
                connect_monitor(self.id, 2, p2_obj)
                game = self.manager.createGame(p1_character, p2_character, p1_name, p2_name)
                self.manager.runGame(game)
                n += 1

            self.gateway.shutdown_callback_server(raise_exception=False)
            self.gateway.shutdown(raise_exception=False)
            sys.stdout.flush()

        return run_a_game

    def run(self, path=FTG_PATH, block=True, n_game=None, **kwargs):
        self.close()
        orig_path = os.path.abspath('.')
        os.chdir(path)
        # change directory to FTG platform
        self._start_server(**kwargs)
        game = self._make_a_game(self.player_1, self.player_2, n_game, **kwargs)
        os.chdir(orig_path)

        if block:
            game()
        else:
            self.game_thread = Thread(target=game)
            self.game_thread.daemon = True
            self.game_thread.start()


def make_csv_logger(path, log_file_):
    # log performance to csv file
    with open(os.path.join(path, log_file_), 'wt') as f:
        f.write('epoch,time,test-score,train-score,loss\n')

    def func_write_to_csv(epoch_, start_time_, testing_score_, training_score_, losses_):
        with open(os.path.join(path, log_file_), 'at') as f:
            f.write('{},{},{},{},{}\n'.format(
                epoch_, time.perf_counter() - start_time_,
                testing_score_.mean(), training_score_.mean(), losses_.mean()))

    return func_write_to_csv


def make_sparkline(title):
    # display score as sparkline (bar chart) using unicode

    Red = lambda t: '\033[31m' + t + '\033[0m'
    Blue = lambda t: '\033[34m' + t + '\033[0m'
    bar = '█▇▆▅▄▃▂▁ ▁▂▃▄▅▆▇█'
    barcount = len(bar) - 1
    xs = list()
    xs.append(1e-6)
    xs.append(-1e-6)

    def func_sparkline_msg(x, max_len=100):
        xs.append(x)
        if len(xs) > max_len:
            del xs[0]
        mx = max([abs(x) for x in xs])
        mn = -mx
        extent = mx - mn
        sparkline = ''.join(Red(bar[int((x - mn) / extent * barcount)]) if x < 0
                            else Blue(bar[int((x - mn) / extent * barcount)]) for x in xs)

        max_len = 10
        mark = ''  # '∧∨△▽▲▼'
        delta = 0.5
        diff = abs(xs[-2] - xs[-1])
        if xs[-1] > xs[-2]:
            mark = Blue('▲')
        elif xs[-1] < xs[-2]:
            mark = Red('▼')
        mark = mark * int(round(diff / delta)) + ' ' * int(max_len - round(diff / delta))
        return '{}: {:5s} {:.5f} {}'.format(title, mark, diff, sparkline)

    return func_sparkline_msg


class PyGameMonitor(object):
    # show input data using pygame surface
    def __init__(self, height, width, debug_queue):
        import pygame
        import skimage.transform

        pygame.init()
        pygame.display.set_caption('FTG3')
        self.display = pygame.display
        self.pixelcopy = pygame.pixelcopy
        self.resize = skimage.transform.resize

        self.debug_port = debug_queue
        self.scale = 3
        self.n_frame = resolution[0]
        self.height = height
        self.width = width
        self.img_buff = np.zeros((self.height, self.n_frame * self.width, 3))
        self.screen = pygame.display.set_mode(
            (self.n_frame * self.scale * self.width, self.scale * self.height))
        self.surf = pygame.Surface(
            (self.n_frame * self.scale * self.width, self.scale * self.height))

    def set_image(self, screen):
        for i in range(self.n_frame):
            ob = screen[i].reshape((self.height, self.width))
            zeros = np.zeros((self.height, self.width))
            my_char = np.where(ob > 0, ob, zeros)
            opp_char = np.where(ob < 0, - ob, zeros)
            self.img_buff[:, i * self.width:(i + 1) * self.width, :] \
                = np.stack([my_char, 0.5 * (my_char + opp_char), opp_char], axis=2)

    def show(self):
        img = self.resize(self.img_buff,
                          (self.scale * self.height, self.n_frame * self.scale * self.width))
        img = (255 * img).astype(np.uint8).swapaxes(0, 1)
        self.pixelcopy.array_to_surface(self.surf, img)
        self.screen.blit(self.surf, (0, 0))
        self.display.flip()

    def run(self):
        while True:
            screen, info = self.debug_port.get()
            self.set_image(screen)
            self.show()
            time.sleep(0.016)


class MatplotlibMonitor(object):
    # show input image state and energy value and output action
    def __init__(self, height, width, debug_queue):
        try:
            import matplotlib.pyplot as plt
            import matplotlib.animation as animation
        except ImportError as exc:
            logger.error('Install matplotlib')
            logger.error(exc)
            exit(-1)

        self.height, self.width = height, width
        self.debug_port = debug_queue

    def run(self):
        channel, height, width = resolution
        display_buffers = [np.zeros((resolution[1], resolution[2])) for _ in range(channel)]
        text_buffer = ['' for _ in range(channel)]
        tick_time = 16.67

        def frames():
            cnt = 0
            screens, info = np.random.random(resolution) * 0.5 - 0.25, dict()

            while True:
                if not self.debug_port.empty():
                    cnt = 0
                    screens, info = self.debug_port.get_nowait()
                else:
                    cnt += 1
                    if cnt > 10:
                        screens, info = np.random.random(resolution) * 0.5 - 0.25, dict()

                yield screens, info
                # time.sleep(0.016)

        fig, axes = plt.subplots(1, 4, figsize=(30, 4), sharey=True)
        fig.suptitle('Vision')
        imgs = list()
        txts = list()
        for i, ax in enumerate(axes):
            ax.set_title('t - {}'.format(channel - i - 1))
            img = ax.imshow(np.zeros((resolution[1], resolution[2]), dtype=np.float32),
                            vmin=-1, vmax=1, cmap='bwr', interpolation='none')
            imgs.append(img)

            text = ax.text(5, 15, '----', fontsize=18)
            txts.append(text)

        def set_data(data):
            screens, info = data
            for i, (screen, disp_buffer, img, text) in enumerate(zip(screens, display_buffers, imgs, txts)):
                disp_buffer[:] = screen[:]
                text_buffer[i] = 'A: {}\nE: {}\nR :{}'.format(
                    info.get('action', 'No data'),
                    info.get('energy', 'No data'),
                    info.get('reward', 'No data'))

                img.set_data(disp_buffer)
                text.set_text(text_buffer[i])

            return imgs + txts

        ani = animation.FuncAnimation(fig, set_data, frames, blit=True, interval=tick_time, repeat=False)
        plt.show()


def make_monitor(kind, env_id=0, player_no=1):

    def func_do_nothing(x):
        return x

    from multiprocessing import Queue
    from multiprocessing import Process

    try:
        import pygame
    except ImportError:
        logger.warning("Can't import pygame")
        return func_do_nothing

    try:
        import skimage.transform
    except ImportError:
        logger.warning("Can't import scikit-image")
        import skimage.transform
        return func_do_nothing

    debug_port = Queue()
    if kind == 'pygame':
        logger.debug('Make PyGame monitor')
        monitor = PyGameMonitor(resolution[1], resolution[2], debug_port)
        thread = Thread(target=monitor.run)
        thread.daemon = True
        thread.start()
    elif kind == 'matplotlib':
        logger.debug('Make Matplotlib monitor')
        monitor = MatplotlibMonitor(resolution[1], resolution[2], debug_port)
        proc = Process(target=monitor.run)
        proc.daemon = True
        proc.start()
    else:
        monitor = None
        return lambda a, b, c: func_do_nothing

    def connect_monitor(eid, pno, player_obj):
        if eid == env_id and pno == player_no and player_obj is not None:
            setattr(player_obj, 'debug_port', monitor.debug_port)
        else:
            if player_obj is None:
                logger.warning("Can't monitor java bot")
            return func_do_nothing

    return connect_monitor


if __name__ == '__main__':

    parser = argparse.ArgumentParser(description="BasicBot example for Visual based Fighting Game AI Platform")
    parser.add_argument('train_or_test', nargs='+', choices=['train', 'test'],
                        help='Select Traning mode or Testing mode')
    parser.add_argument('--max-epoch', default=None, type=int, help='Max epochs for training')
    parser.add_argument('-ts', '--training-step', type=int, default=50000, help='set how many frame use for training')
    parser.add_argument('-es', '--evaluation-sec', type=int, default=180,
                        help='set how many time(seconds) use for evaluation (testing) without learning')
    parser.add_argument('-o', '--opponent', default='SandBag',
                        help='''opponent name
  - Python bot: opponent class name and filename shoud be same and it placed in same directory
  - Java bot: should be add character type like 'MctsAi:LUD', however it is not test enough''')
    parser.add_argument('-n', '--n-cpu', default=3, type=int,
                        help='set how many games launch, it should lower than number of CPU cores, '
                             'also considering limits of other computing resources')
    parser.add_argument('-p', '--path', default='.',
                        help='create this directory if it is not exists '
                             'and write csv file as a log and store networks parameters')
    parser.add_argument('--port', default=6000, type=int, help='help')
    parser.add_argument('-r', '--render', default='single', choices=['none', 'single', 'all'], help='game server port')
    parser.add_argument('-v, --verbose', dest='verbose', action='store_true', help='show debug level message')
    parser.set_defaults(verbose=False)
    parser.add_argument('-m', '--monitor', default='none',
                        choices=['none', 'pygame', 'matplotlib'],
                        help='show current input screen data and reward value and out action, '
                             'two types supported pygame and matplotlib')
    parser.add_argument('--starts-with-energy',dest='starts_with_energy', action='store_true',
                        help='get max energy for each character when start a game')
    parser.set_defaults(starts_with_energy=False)
    parser.add_argument('--disable-tqdm', dest='use_tqdm', action='store_false', help='disable tqdm')
    parser.set_defaults(use_tqdm=True)
    args = parser.parse_args()

    logger.info('Launch {:d} environment(s)'.format(args.n_cpu))
    logger.info('{} vs. {}'.format(bot_name, args.opponent))

    if not args.use_tqdm:
        logger.info('Disable tqdm.trange')

        def _trange(*args, **kwargs):
            return range(*args)
        # override tqdm trange
        globals()['trange'] = _trange

    opponent = args.opponent
    if re.search("^[a-z0-9]+:(ZEN|LUD|GARNET)$", args.opponent, re.I):
        # for Java bot: eg. 'MctsAi:LUD'
        logger.info("Set java opponent: {}".format(args.opponent))
    else:
        # for python bot: bot's class name and file name should same,
        # the file should be place where this script can import
        try:
            mod = importlib.import_module(args.opponent)
            opponent = getattr(mod, args.opponent)
            logger.info("Set python opponent: {}".format(args.opponent))
        except ImportError:
            logger.error("Can't import {}".format(args.opponent))
            sys.exit(-1)

    root_path = args.path
    logger.info('Set path: {}'.format(root_path))
    os.makedirs(root_path, exist_ok=True)

    step_per_training = args.training_step
    seconds_per_testing = args.evaluation_sec

    if args.verbose:
        logger.setLevel(logging.DEBUG)
        logger.debug('Verbose mode on')

    logger.info('** Start {}ING ** '.format(args.train_or_test[0].upper()))
    # Train start
    if 'train' in args.train_or_test:
        logger.debug('Open {} as logfile to write'.format(os.path.join(root_path, log_file)))
        write_to_csv = make_csv_logger(root_path, log_file)

    sparkline_gen = make_sparkline('Scores')
    ai_monitor = make_monitor(args.monitor, env_id=0, player_no=1)

    envs = list()
    logger.info('Args: {}'.format(args))
    if 'train' in args.train_or_test:
        for n in range(args.n_cpu):
            disable_window = not (args.render == 'on' or args.render == 'single' and n == 0)
            env = FTGEnv(n, BasicBot, opponent, port=args.port + n,
                         inverted_player=2,
                         disable_window=disable_window,
                         starts_with_energy=args.starts_with_energy,
                         train='train' in args.train_or_test, verbose=False)
            env.run(block=False, ai_monitor=ai_monitor)
            envs.append(env)

        while memory.size < batch_size:
            time.sleep(0.016 * frames)
            pass

        epoch = 0
        start_time = time.perf_counter()
        highest_score = -1e-10

        while True:
            Coach().training(epsilon)
            q_values = np.zeros(step_per_training)
            losses = np.zeros(step_per_training)
            for step in trange(step_per_training, desc='Training {}'.format(epoch)):
                max_q, loss = Coach().learn()
                q_values[step] = max_q.mean()
                losses[step] = loss
            training_score = Coach().get_rewards_stat()

            Coach().testing()
            for step in trange(seconds_per_testing, desc='Testing {}'.format(epoch)):
                time.sleep(1)
            testing_score = Coach().get_rewards_stat()

            logger.info('Epoch: {} Testing: {:.3f} ± {:.3f} Training: {:.3f} ± {:.3f} Loss: {:.9f} Q: {:.3f} ε: {:.5f}'.format(
                epoch, testing_score.mean(), testing_score.std(), training_score.mean(), training_score.std(),
                losses.mean(), q_values.mean(), epsilon))
            logger.info('Epoch: {} {}'.format(epoch, sparkline_gen(testing_score.mean())))

            write_to_csv(epoch, start_time, testing_score, training_score, losses)

            if testing_score.mean() > highest_score:
                highest_score = testing_score.mean()
                Coach().save(model_file, hint=highest_score)
                logger.info('New highest score: {}'.format(highest_score))

            # reduce epsilon to minimum (0.1) until half of the max_epoch
            epsilon_delta = 0.05 if args.max_epoch is None else 1.0 / args.max_epoch * 2
            epsilon = max(0.1, epsilon - epsilon_delta)
            epoch += 1
            if args.max_epoch is not None and epoch > args.max_epoch:
                break

    if 'test' in args.train_or_test:
        disable_window = not (args.render == 'on' or args.render == 'single')
        env = FTGEnv(0, BasicBot, opponent, port=args.port,
                     inverted_player=2,
                     disable_window=disable_window,
                     starts_with_energy=args.starts_with_energy,
                     train='train' in args.train_or_test, verbose=False)
        env.run(block=True, ai_monitor=ai_monitor)
        envs.append(env)

    logger.info('** Stop {}ING ** '.format(args.train_or_test[0].upper()))
    logger.info('Press ctrl + c many times to exit')
